CW_MAX_CHANNELS = 10;
CW_MAX_CHANNEL_ROSTER = 9;

-- This comes from Choonster @ Frostmourne: http://us.battle.net/wow/en/forum/topic/4044466747
CW_onlineMatchString = ERR_FRIEND_ONLINE_SS:format(".+", "(.+)%"):gsub("%[","%%[");
CW_offlineMatchString = ERR_FRIEND_OFFLINE_S:format("(.+)");

-- Inject a count field into the GuildMicroButton
local CW_GCText = GuildMicroButtonTabard:CreateFontString("GuildMicroButtonTabardCount", "OVERLAY", "NumberFontNormal");
CW_GCText:SetPoint("BOTTOMRIGHT", -4, 4);
CW_GCText:SetJustifyH("RIGHT");
CW_GCText:Hide();

-- local
CW_Channels = {}; -- Array of channel structures allowing much reuse.

--[[
	Guild Support functions
--]]
function ChatWho_GuildCountUpdate()
	if IsInGuild() then
		local totalMembers, onlineMembers, onlineAndMobileMembers = GetNumGuildMembers();
		CW_GCText:SetText(onlineMembers);
		CW_GCText:Show();
	else
		CW_GCText:Hide();
	end
end

--[[
	Channel Support functions
--]]
CW_WorkQueue = {};
CW_Updating = false;
CW_Paused = false;
function ChatWho_QueueChannelInfo( channelID )
	if ( not CW_Channels[channelID].used ) then  return;  end

	for i, v in pairs(CW_WorkQueue) do
		if (v == channelID) then
--			print(">>Ignoring "..channelID.." Already in queue");
			return;
		end;
	end

	if( not CW_Updating and not CW_Paused) then
		ChatWho_ChannelQueryServer( channelID );
	else
		table.insert(CW_WorkQueue, channelID);
	end
	-- Regardless of the previous check. We should always be waiting for an update after this
	CW_Updating = true;
--	print(format(">> ChatWho_QueueChannelInfo: %d, Updating? %s Queue length? %d", channelID, tostring(CW_Updating), #CW_WorkQueue));
end

function ChatWho_HandleChannelInfo( displayID )
	-- We may get server responses we didn't request.
	-- We are A-OK with this. The info is still good, and possibly more up to date.
	local name, header, collapsed, channelNumber, count, active, category, voiceEnabled, voiceActive = GetChannelDisplayInfo( displayID );

	if ( category and category == "CHANNEL_CATEGORY_CUSTOM") then
		CW_Channels[channelNumber].name = name;
		CW_Channels[channelNumber].displayID = displayID;
		CW_Channels[channelNumber].count = count;
		CW_Channels[channelNumber].used = true;
		
		for i=1, min(count, CW_MAX_CHANNEL_ROSTER) do
			CW_Channels[channelNumber].roster[i] = GetChannelRosterInfo(displayID, i);
		end
--		print("Loaded roster for "..name.."("..count..")");
	else
		CW_Channels[channelNumber].used = false;
	end
	
	ChatWho_UpdateButtons();
	ChatWho_ChannelQueryNext();
end

function ChatWho_ChannelQueryNext()
--	print(">> CW_CQN: "..tostring(CW_Updating)..", "..tostring(CW_Paused));
	if ( CW_Updating and not CW_Paused) then
		local nextChannelID = table.remove(CW_WorkQueue, 1);
		if ( nextChannelID ) then
			ChatWho_ChannelQueryServer( nextChannelID );
		else
--			print(">> Queue done!");
			CW_Updating = false;
		end
	end
end

CW_LastID = 0;
function ChatWho_ChannelQueryServer( queryID )
	if ( not CW_Channels[queryID].used ) then
		-- Just in case our Queue was off
		-- or a channel got queued and deleted while we were paused.
		ChatWho_ChannelQueryNext();
	end
	
	-- If the last query was asked recently. Toggle the question quickly to ask the server again
	local queryDisplayID = CW_Channels[queryID].displayID;
	
	if ( CW_LastID < 1 or queryDisplayID == CW_LastID ) then
		for i=1, GetNumDisplayChannels() do
			local name, header, collapsed, channelNumber, count, active, category, voiceEnabled, voiceActive = GetChannelDisplayInfo(i);
			if ( channelNumber and i ~= queryDisplayID) then
				SetSelectedDisplayChannel( i );
				break;
			end
		end
	end
	
	SetSelectedDisplayChannel( queryDisplayID );
	CW_LastID = queryDisplayID;
end

function ChatWho_ChannelPause()
	CW_Paused = true;
end

function ChatWho_ChannelUnpause()
	if ( not CW_Paused ) then return; end
	CW_Paused = false;
	ChatWho_ChannelQueryNext();
end

-- CHAT_MSG_CHANNEL_NOTICE - For player leaving/joining a channel. Superseded by CHANNEL_UI_UPDATE.
function ChatWho_SetButtonSide(chatFrame, buttonSide)
	ChatWhoButton1:ClearAllPoints();
	if (buttonSide == "right") then		
		ChatWhoButton1:SetPoint("TOPLEFT", FriendsMicroButton, "TOPRIGHT", -2, 0);
	else
		ChatWhoButton1:SetPoint("TOPRIGHT", FriendsMicroButton, "TOPLEFT", 2, 0);
	end
end

function ChatWho_OnLoad(self, ...)
	self:RegisterEvent("GUILD_ROSTER_UPDATE");
	self:RegisterEvent("PLAYER_GUILD_UPDATE");
	self:RegisterEvent("CHANNEL_ROSTER_UPDATE");
	self:RegisterEvent("PLAYER_ENTERING_WORLD");
	self:RegisterEvent("CHAT_MSG_CHANNEL_JOIN");
	self:RegisterEvent("CHAT_MSG_CHANNEL_LEAVE");
	self:RegisterEvent("CHAT_MSG_SYSTEM");
	self:RegisterEvent("CHANNEL_UI_UPDATE");
	
	-- The following controls the fixed offset of the chat frame: (32 is the DEFAULT, from UIParent.lua)
	-- And default clamping is 	:SetClampRectInsets(-35, 35, 26, -50);
	UIPARENT_MANAGED_FRAME_POSITIONS[DEFAULT_CHAT_FRAME:GetName()].xOffset = 60;
	DEFAULT_CHAT_FRAME:SetClampRectInsets(-62, 62, 26, -50);
	hooksecurefunc("FCF_SetButtonSide", ChatWho_SetButtonSide);
--	hooksecurefunc("GuildRoster", function() print("GuildRoster() Called\n", debugstack(3,3,2)) end)
	ChannelFrame:HookScript("OnShow", ChatWho_ChannelPause);
	ChannelFrame:HookScript("OnHide", ChatWho_ChannelUnpause);
	
	for i=2, CW_MAX_CHANNELS do
		local button = CreateFrame("Button", "ChatWhoButton"..i, UIParent, "ChatWhoButtonTemplate");
		button:SetPoint("TOPRIGHT", _G["ChatWhoButton"..(i-1)], "BOTTOMRIGHT", 0, -4);
		button:GetNormalTexture():SetDesaturated(true);
	end
	ChatWhoButton1:GetNormalTexture():SetDesaturated(true);
	
	-- Channels are numbered as per warcraft client
	for i=1, CW_MAX_CHANNELS do
		CW_Channels[i] = {
			name = nil,
			displayID = nil, -- ID used by the ChannelFrame UI, this is channel number + some value 3 or higher
			used = false,
			count = 0,
			roster = { UnitName("Player") },
			tag = "CHANNEL"..i
		};
	end
end

function ChatWho_OnEvent(self, event, ...)
	local arg1, arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, arg10, arg11, arg12, arg13, arg14 = ...;

	-- This missed updating when a player joined the guild
	if ( event == "GUILD_ROSTER_UPDATE" or event == "PLAYER_GUILD_UPDATE" ) then
		ChatWho_GuildCountUpdate();

	elseif ( event == "CHANNEL_ROSTER_UPDATE" ) then
--		print(">> CHANNEL_ROSTER_UPDATE "..arg1);
		ChatWho_HandleChannelInfo( arg1 );

	elseif ( event == "CHAT_MSG_SYSTEM" ) then
		-- Other things may trigger the GuildRoster() call, but just in case..
--		print(">> "..event..":", ">>"..arg1.."<<", arg2, arg3, arg4, arg5, arg6, arg7, arg8, arg9, arg10, arg11, arg12, arg13, arg14 )
--		CW_STR1 = arg1; -- Debug str for me to dump
		if ( IsInGuild() and (arg1:match(CW_onlineMatchString) or arg1:match(CW_offlineMatchString))) then
--			print(">> GuildRoster() fired");
			GuildRoster();
		end

	elseif ( strsub(event, 1, 8) == "CHAT_MSG" ) then
		-- Channel Join/Leave
--		print(">> "..event..":", arg1, arg2, arg3, arg4, arg5, arg6, arg7, ">>"..arg8.."<<", arg9, arg10, arg11, arg12, arg13, arg14 )
		if (CW_Channels[arg8].used) then
			ChatWho_QueueChannelInfo( arg8 );
			if (event == "CHAT_MSG_CHANNEL_JOIN" and CW_Channels[arg8].used) then
				CW_Channels[arg8].button.playerJoined:Play();
			end
		end

	elseif ( event == "PLAYER_ENTERING_WORLD" or event == "CHANNEL_UI_UPDATE" ) then
		-- Occurs before channels are joined,
		-- but we don't rejoin channels when we /reload
		-- So this helps

		-- Try not to overdue calls to this.
		-- For instance Login is 1 PLAYER_ENTERING_WORLD and 10+ CHANNEL_UI_UPDATE 
		if (not CW_Updating or CW_Paused) then
			ChatWho_UpdateAllChannels( );
		end
	end
end

local firstTime = true;
function ChatWho_UpdateAllChannels()
	-- We might not be told of removed channels. Sweep and correct
	for i=1, CW_MAX_CHANNELS do
		CW_Channels[i].used = false;
	end

	-- If the last Display entry is a header, force it to be open.
	local updateButtons = true;
	if( select(2, GetChannelDisplayInfo( GetNumDisplayChannels() )) ) then
		ExpandChannelHeader( GetNumDisplayChannels() );
		updateButtons = false; -- Skip updating, the subsequent call for a UI Update will get them
	end
	
	-- ChannelFrame sometimes gets confused, prevent it from showing its DropDown
	ChannelListDropDown.clicked = nil;

	local withQuery = firstTime;
	for i=1, GetNumDisplayChannels() do
		local name, header, collapsed, channelNumber, count, active, category, voiceEnabled, voiceActive = GetChannelDisplayInfo(i);
		if ( category and category == "CHANNEL_CATEGORY_CUSTOM") then
			CW_Channels[channelNumber].name = name;
			CW_Channels[channelNumber].displayID = i;
			if ( count ) then
				CW_Channels[channelNumber].count = count;
			else
				withQuery = true;
			end
			CW_Channels[channelNumber].used = true;

			if ( withQuery ) then
				ChatWho_QueueChannelInfo( channelNumber );
			end
		end
	end
	
	if (updateButtons) then
		ChatWho_UpdateButtons();
	end
	firstTime = false;
end

function ChatWho_UpdateButtons()
	local i = 1;
	local info;

	for k=1, CW_MAX_CHANNELS do
		if ( CW_Channels[k].used ) then
			info = ChatTypeInfo[CW_Channels[k].tag];
			local button = _G["ChatWhoButton"..i];
			button:SetID(k);
			button.tint:SetTexture(info.r, info.g, info.b);
			if (CW_Channels[k].count) then _G["ChatWhoButton"..i.."Count"]:SetText(CW_Channels[k].count); end
			button:Show();
			CW_Channels[k].button = button;
			i = i + 1
		end
	end
	
	for k = i, CW_MAX_CHANNELS do
		_G["ChatWhoButton"..k]:Hide();
	end
end
